import re
from collections import namedtuple
from urllib.parse import unquote

from django.db.models import QuerySet
from django.utils.translation import gettext as _
from rest_framework.serializers import ValidationError

from rest_framework_filters.utils import lookahead

# originally based on: https://regex101.com/r/5rPycz/1
# current iteration: https://regex101.com/r/5rPycz/3
# special thanks to @JohnDoe2 on the #regex IRC channel!
# matches groups of "<negate>(<encoded querystring>)<set op>"
COMPLEX_OP_RE = re.compile(r'()\(([^)]+)\)([^(]*?(?=\())?')
COMPLEX_OP_NEG_RE = re.compile(r'(~?)\(([^)]+)\)([^(]*?(?=~\(|\())?')
COMPLEX_OPERATORS = {
    '&': QuerySet.__and__,
    '|': QuerySet.__or__,
}

ComplexOp = namedtuple('ComplexOp', ['querystring', 'negate', 'op'])


def decode_complex_ops(encoded_querystring, operators=None, negation=True):
    """Decode the complex encoded querysting into a list of complex operations.

    .. code-block:: python

        # unencoded query: (a=1) & (b=2) | ~(c=3)
        >>> s = '%28a%253D1%29%20%26%20%28b%253D2%29%20%7C%20%7E%28c%253D3%29'
        >>> decode_querystring_ops(s)
        [
            ('a=1', False, QuerySet.__and__),
            ('b=2', False, QuerySet.__or__),
            ('c=3', True, None),
        ]

    Args:
        encoded_querystring: The encoded querystring.
        operators: A map of {operator symbols: queryset operations}. Defaults to the
            ``COMPLEX_OPERATIONS`` mapping.
        negation: Whether to parse negation.

    Returns:
        A list of ``(querystring, negate, op)`` tuples that represent the operations.

    Raises:
        ValidationError: Raised under the following conditions:
            - the individual querystrings are not wrapped in parentheses
            - the set operators do not match the provided `operators`
            - there is trailing content after the ending querysting
    """
    complex_op_re = COMPLEX_OP_NEG_RE if negation else COMPLEX_OP_RE
    if operators is None:
        operators = COMPLEX_OPERATORS

    # decode into: (a%3D1) & (b%3D2) | ~(c%3D3)
    decoded_querystring = unquote(encoded_querystring)
    matches = list(complex_op_re.finditer(decoded_querystring))

    if not matches:
        msg = _("Unable to parse querystring. Decoded: '%(decoded)s'.")
        raise ValidationError(msg % {'decoded': decoded_querystring})

    results, errors = [], []
    for match, has_next in lookahead(matches):
        negate, querystring, op = match.groups()

        negate = negate == '~'
        querystring = unquote(querystring)
        op_func = operators.get(op.strip()) if op else None
        if op_func is None and has_next:
            msg = _("Invalid querystring operator. Matched: '%(op)s'.")
            errors.append(msg % {'op': op})

        results.append(ComplexOp(querystring, negate, op_func))

    msg = _("Ending querystring must not have trailing characters. Matched: '%(chars)s'.")
    trailing_chars = decoded_querystring[matches[-1].end():]
    if trailing_chars:
        errors.append(msg % {'chars': trailing_chars})

    if errors:
        raise ValidationError(errors)

    return results


def combine_complex_queryset(querysets, complex_ops, negation=True):
    # Negate querysets
    for queryset, op in zip(querysets, complex_ops):
        if negation and op.negate:
            queryset.query.where.negate()

    # Combine querysets
    combined = querysets[0]
    for queryset, op in zip(querysets[1:], complex_ops[:-1]):
        combined = op.op(combined, queryset)

    return combined



import warnings

from django.utils.module_loading import import_string
from django_filters.rest_framework.filters import *  # noqa
from django_filters.rest_framework.filters import (
    ModelChoiceFilter, ModelMultipleChoiceFilter,
)

ALL_LOOKUPS = '__all__'


class AutoFilter:
    """A placeholder class that enables generating multiple per-lookup filters.

    This is a declarative alternative to the ``Meta.fields`` dict syntax, and
    the below are functionally equivalent:

    .. code-block:: python

        class PersonFilter(filters.FilterSet):
            name = AutoFilter(lookups=['exact', 'contains'])

            class Meta:
                model = Person
                fields = []

        class PersonFilter(filters.FilterSet):
            class Meta:
                model = Person
                fields = {'name': ['exact', 'contains']}

    Due to its declarative nature, an ``AutoFilter`` allows for paramater name
    aliasing for its generated filters. e.g.,

    .. code-block:: python

        class BlogFilter(filters.FilterSet):
            title = AutoFilter(field_name='name', lookups=['contains'])

    The above generates a ``title__contains`` filter for the ``name`` field
    intstead of ``name__contains``. This is not possible with ``Meta.fields``,
    since the key must match the model field name.

    Note that ``AutoFilter`` is a filter-like placeholder and is not present in
    the ``FilterSet.filters``. In the above example, ``title`` would not be
    filterable. However, an ``AutoFilter`` is typically replaced by a generated
    ``exact`` filter of the same name, which enables filtering by that param.
    """

    creation_counter = 0

    def __init__(self, field_name=None, *, lookups=None):
        self.field_name = field_name
        self.lookups = lookups or []

        self.creation_counter = AutoFilter.creation_counter
        AutoFilter.creation_counter += 1


class BaseRelatedFilter:

    def __init__(self, filterset, *args, lookups=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.filterset = filterset
        self.lookups = lookups or []

    def bind_filterset(self, filterset):
        """Bind a filterset class to the filter instance.

        This class is used for relative imports. Only the first bound class is used as
        filterset inheritance might otherwise break these relative import paths.

        This is also necessary to allow ``.filterset`` to be resolved during FilterSet
        class creation time, instead of during initialization.

        Args:
            filterset: The filterset to bind
        """
        if not hasattr(self, 'bound_filterset'):
            self.bound_filterset = filterset

    def filterset():
        def fget(self):
            if isinstance(self._filterset, str):
                try:
                    # Assume absolute import path
                    self._filterset = import_string(self._filterset)
                except ImportError:
                    # Fallback to building import path relative to bind class
                    path = '.'.join([self.bound_filterset.__module__, self._filterset])
                    self._filterset = import_string(path)
            return self._filterset

        def fset(self, value):
            self._filterset = value

        return locals()
    filterset = property(**filterset())

    def get_queryset(self, request):
        queryset = super(BaseRelatedFilter, self).get_queryset(request)
        assert queryset is not None, \
            "Expected `.get_queryset()` for related filter '%s.%s' to " \
            "return a `QuerySet`, but got `None`." \
            % (self.parent.__class__.__name__, self.field_name)
        return queryset


class RelatedFilter(BaseRelatedFilter, ModelChoiceFilter):
    """A ``ModelChoiceFilter`` that enables filtering across relationships.

    Take the following example:

        class ManagerFilter(filters.FilterSet):
            class Meta:
                model = Manager
                fields = {'name': ['exact', 'in', 'startswith']}

        class DepartmentFilter(filters.FilterSet):
            manager = RelatedFilter(ManagerFilter, queryset=managers)

            class Meta:
                model = Department
                fields = {'name': ['exact', 'in', 'startswith']}

    In the above, the ``DepartmentFilter`` can traverse the ``manager``
    relationship with the ``__`` lookup seperator, accessing the filters of the
    ``ManagerFilter`` class. For example, the above would enable calls like:

        /api/managers?name=john%20doe
        /api/departments?manager__name=john%20doe

    Related filters function similarly to auto filters in that they can generate
    per-lookup filters. However, unlike auto filters, related filters are
    functional and not just placeholders. They will not be replaced by a
    generated ``exact`` filter.

    Attributes:
        filterset: The ``FilterSet`` that is traversed by this relationship.
            May be a class, an absolute import path, or the name of a class
            located in the same module as the origin filterset.
        lookups: A list of lookups to generate per-lookup filters for. This
            functions similarly to the ``AutoFilter.lookups`` argument.
    """


class RelatedMultipleFilter(BaseRelatedFilter, ModelMultipleChoiceFilter):
    """A ``ModelMultipleChoiceFilter`` variant of ``RelatedFilter``."""


class AllLookupsFilter(AutoFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, lookups=ALL_LOOKUPS, **kwargs)
        warnings.warn(
            "`AllLookupsFilter()` has been deprecated in favor of "
            "`AutoFilter(lookups='__all__')`.",
            DeprecationWarning,
            stacklevel=2,
        )



from django.db.models.constants import LOOKUP_SEP
from django.db.models.expressions import Expression
from django.db.models.lookups import Transform


def lookups_for_field(model_field):
    """Generate a list of all possible lookup expressions for a model field.

    Args:
        model_field: The model field to inspect.

    Returns:
        A list of lookups for the given ``model_field``.
    """
    lookups = []

    for expr, lookup in model_field.get_lookups().items():
        if issubclass(lookup, Transform):
            transform = lookup(Expression(model_field))
            lookups += [
                LOOKUP_SEP.join([expr, sub_expr]) for sub_expr
                in lookups_for_transform(transform)
            ]

        else:
            lookups.append(expr)

    return lookups


def lookups_for_transform(transform):
    """Generate a list of subsequent lookup expressions for a transform.

    Note:
    Infinite transform recursion is only prevented when the subsequent and passed in
    transforms are the same class. For example, the ``Unaccent`` transform from
    ``django.contrib.postgres``.
    There is no cycle detection across multiple transforms. For example, ``a__b__a__b``
    would continue to recurse. However, this is not currently a problem (no builtin
    transforms exhibit this behavior).

    Args:
        transform: The transform to inspect.

    Returns:
        A list of lookups for the given ``transform``.
    """
    lookups = []

    for expr, lookup in transform.output_field.get_lookups().items():
        if issubclass(lookup, Transform):

            # type match indicates recursion.
            if type(transform) == lookup:
                continue

            sub_transform = lookup(transform)
            lookups += [
                LOOKUP_SEP.join([expr, sub_expr]) for sub_expr
                in lookups_for_transform(sub_transform)
            ]

        else:
            lookups.append(expr)

    return lookups


def lookahead(iterable):
    it = iter(iterable)
    try:
        current = next(it)
    except StopIteration:
        return

    for value in it:
        yield current, True
        current = value
    yield current, False



from contextlib import contextmanager

from django.http import QueryDict
from django_filters import compat
from django_filters.rest_framework import backends
from rest_framework.exceptions import ValidationError

from .complex_ops import combine_complex_queryset, decode_complex_ops
from .filterset import FilterSet


class RestFrameworkFilterBackend(backends.DjangoFilterBackend):
    filterset_base = FilterSet

    @property
    def template(self):
        if compat.is_crispy():
            return 'rest_framework_filters/crispy_form.html'
        return 'rest_framework_filters/form.html'

    @contextmanager
    def patch_for_rendering(self, request):
        # Patch ``.get_filterset_class()`` so the resulting filterset does not perform
        # filter expansion during form rendering.
        original = self.get_filterset_class

        def get_filterset_class(view, queryset=None):
            filterset_class = original(view, queryset)

            # Don't break if filterset_class is not provided
            if filterset_class is None:
                return None

            # django-filter compatibility
            if issubclass(filterset_class, FilterSet):
                filterset_class = filterset_class.disable_subset(depth=1)

            return filterset_class

        self.get_filterset_class = get_filterset_class
        try:
            yield
        finally:
            self.get_filterset_class = original

    def to_html(self, request, queryset, view):
        # Patching the behavior of ``.get_filterset_class()`` in this method allows us
        # to avoid maintenance issues with code duplication.
        with self.patch_for_rendering(request):
            return super().to_html(request, queryset, view)


class ComplexFilterBackend(RestFrameworkFilterBackend):
    complex_filter_param = 'filters'
    operators = None
    negation = True

    def filter_queryset(self, request, queryset, view):
        if self.complex_filter_param not in request.query_params:
            return super().filter_queryset(request, queryset, view)

        # Decode the set of complex operations
        encoded_querystring = request.query_params[self.complex_filter_param]
        try:
            complex_ops = decode_complex_ops(
                encoded_querystring,
                self.operators,
                self.negation,
            )
        except ValidationError as exc:
            raise ValidationError({self.complex_filter_param: exc.detail})

        # Collect the individual filtered querysets
        querystrings = [op.querystring for op in complex_ops]
        try:
            querysets = self.get_filtered_querysets(querystrings, request, queryset, view)
        except ValidationError as exc:
            raise ValidationError({self.complex_filter_param: exc.detail})

        return combine_complex_queryset(querysets, complex_ops)

    def get_filtered_querysets(self, querystrings, request, queryset, view):
        original_GET = request._request.GET

        querysets, errors = [], {}
        for qs in querystrings:
            request._request.GET = QueryDict(qs)
            try:
                result = super().filter_queryset(request, queryset, view)
                querysets.append(result)
            except ValidationError as exc:
                errors[qs] = exc.detail
            finally:
                request._request.GET = original_GET

        if errors:
            raise ValidationError(errors)
        return querysets



import copy
from collections import OrderedDict

from django.db.models.constants import LOOKUP_SEP
from django_filters import filterset, rest_framework
from django_filters.utils import get_model_field

from . import filters, utils


def related(filterset, filter_name):
    # Return a related filter_name, using the filterset relationship if present.
    if not filterset.relationship:
        return filter_name
    return LOOKUP_SEP.join([filterset.relationship, filter_name])


class FilterSetMetaclass(filterset.FilterSetMetaclass):
    def __new__(cls, name, bases, attrs):
        attrs['auto_filters'] = cls.get_auto_filters(bases, attrs)

        new_class = super().__new__(cls, name, bases, attrs)

        new_class.related_filters = OrderedDict([
            (name, f) for name, f in new_class.declared_filters.items()
            if isinstance(f, filters.BaseRelatedFilter)])

        # See: :meth:`rest_framework_filters.filters.RelatedFilter.bind`
        for f in new_class.related_filters.values():
            f.bind_filterset(new_class)

        # Only expand when model is defined. Model may be undefined for mixins.
        if new_class._meta.model is not None:
            for name, f in new_class.auto_filters.items():
                expanded = cls.expand_auto_filter(new_class, name, f)
                new_class.base_filters.update(expanded)

            for name, f in new_class.related_filters.items():
                expanded = cls.expand_auto_filter(new_class, name, f)
                new_class.base_filters.update(expanded)

        return new_class

    @classmethod
    def get_auto_filters(cls, bases, attrs):
        # Auto filters are specially handled since they aren't an actual filter
        # subclass and aren't handled by the declared filter machinery. Note
        # that this is a nearly identical copy of `get_declared_filters`.
        auto_filters = [
            (filter_name, attrs.pop(filter_name))
            for filter_name, obj in list(attrs.items())
            if isinstance(obj, filters.AutoFilter)
        ]

        # Default the `filter.field_name` to the attribute name on the filterset
        for filter_name, f in auto_filters:
            if getattr(f, 'field_name', None) is None:
                f.field_name = filter_name

        auto_filters.sort(key=lambda x: x[1].creation_counter)

        # merge auto filters from base classes
        for base in reversed(bases):
            if hasattr(base, 'auto_filters'):
                auto_filters = [
                    (name, f) for name, f
                    in base.auto_filters.items()
                    if name not in attrs
                ] + auto_filters

        return OrderedDict(auto_filters)

    @classmethod
    def expand_auto_filter(cls, new_class, filter_name, f):
        """Resolve an ``AutoFilter`` into its per-lookup filters.

        This method name is slightly inaccurate since it handles both
        :class:`rest_framework_filters.filters.AutoFilter` and
        :class:`rest_framework_filters.filters.BaseRelatedFilter`, as well as
        their subclasses, which all support per-lookup filter generation.

        Args:
            new_class: The ``FilterSet`` class to generate filters for.
            filter_name: The attribute name of the filter on the ``FilterSet``.
            f: The filter instance.

        Returns:
            A named map of generated filter objects.
        """
        expanded = OrderedDict()

        # get reference to opts/declared filters so originals aren't modified
        orig_meta, orig_declared = new_class._meta, new_class.declared_filters
        new_class._meta = copy.deepcopy(new_class._meta)
        new_class.declared_filters = {}

        # Use meta.fields to generate auto filters
        new_class._meta.fields = {f.field_name: f.lookups or []}
        for gen_name, gen_f in new_class.get_filters().items():
            # get_filters() generates param names from the model field name, so
            # replace the field name with the param name from the filerset
            gen_name = gen_name.replace(f.field_name, filter_name, 1)

            # do not overwrite declared filters
            if gen_name not in orig_declared:
                expanded[gen_name] = gen_f

        # restore reference to opts/declared filters
        new_class._meta, new_class.declared_filters = orig_meta, orig_declared

        return expanded


class SubsetDisabledMixin:
    """Disable filter subsetting (see: :meth:`FilterSet.disable_subset`)."""

    @classmethod
    def get_filter_subset(cls, params, rel=None):
        return cls.base_filters


class FilterSet(rest_framework.FilterSet, metaclass=FilterSetMetaclass):

    def __init__(self, data=None, queryset=None, *, relationship=None, **kwargs):
        self.base_filters = self.get_filter_subset(data or {}, relationship)

        super().__init__(data, queryset, **kwargs)

        self.relationship = relationship
        self.related_filtersets = self.get_related_filtersets()
        self.filters = self.get_request_filters()

    @classmethod
    def get_fields(cls):
        # Extend the 'Meta.fields' dict syntax to allow '__all__' field lookups.
        fields = super(FilterSet, cls).get_fields()

        for name, lookups in fields.items():
            if lookups == filters.ALL_LOOKUPS:
                field = get_model_field(cls._meta.model, name)
                if field is not None:
                    fields[name] = utils.lookups_for_field(field)
                else:
                    # FilterSet will handle invalid name
                    fields[name] = []

        return fields

    @classmethod
    def get_filter_subset(cls, params, rel=None):
        """Get the subset of filters that should be initialized by this filterset class.

        A filterset may have a large number of filters, and selecting the subset based on
        the request ``params`` minimizes the cost of initialization by reducing the total
        number of expensive deepcopy operations. See :meth:`.get_param_filter_name()`
        for a better understanding of how the filter names are resolved.

        Args:
            params: The request's query params.
            rel (str, optional): The relationship the ``params`` are resolved against.

        Returns:
            Mapping of the resolved ``{filter names: filter instances}``.
        """
        # Determine names of filters from query params and remove empty values.
        # param names that traverse relations are translated to just the local
        # filter names. eg, `author__username` => `author`. Empty values are
        # removed, as they indicate an unknown field eg, author__foobar__isnull
        filter_names = {cls.get_param_filter_name(param, rel) for param in params}
        filter_names = {f for f in filter_names if f is not None}
        return OrderedDict(
            (k, v) for k, v in cls.base_filters.items() if k in filter_names
        )

    @classmethod
    def disable_subset(cls, *, depth=0):
        """Disable filter subsetting, allowing a form to render the complete filterset.

        Note that this decreases performance and should only be used when rendering a
        form, such as with DRF's browsable API.

        Args:
            depth (int, optional): Disable related filterset subsetting to this depth.
                The form should not render rels beyond this depth.

        Returns:
            This filterset class with subset disabling mixed in.
        """
        if not issubclass(cls, SubsetDisabledMixin):
            cls = type('SubsetDisabled%s' % cls.__name__,
                       (SubsetDisabledMixin, cls), {})

        # recursively disable subset for related filtersets
        if depth > 0:
            # shallow copy to prevent modifying original `base_filters`
            cls.base_filters = cls.base_filters.copy()

            # deepcopy RelateFilter to prevent modifying original `.filterset`
            for name in cls.related_filters:
                f = copy.deepcopy(cls.base_filters[name])
                f.filterset = f.filterset.disable_subset(depth=depth - 1)
                cls.base_filters[name] = f

        return cls

    @classmethod
    def get_param_filter_name(cls, param, rel=None):
        """Resolve a query parameter name into a filter name.

        This is primarily used to resolve query parameters respective to their base or
        related filtersets. e.g.,

        .. code-block:: python

            # regular attribute filters
            >>> FilterSet.get_param_filter_name('email')
            'email'

            # exclusion filters
            >>> FilterSet.get_param_filter_name('email!')
            'email'

            # related filters
            >>> FilterSet.get_param_filter_name('author__email')
            'author'

            # attribute filters based on relationship
            >>> FilterSet.get_param_filter_name('author__email', rel='author')
            'email'

        Args:
            param (str): The query paramater.
            rel (str, optional): The relationship a ``param`` is resolved against.

        Returns:
            The parameter name relative to the relationship.
        """
        # check for empty param
        if not param:
            return param

        # check that param is not the rel.
        if param == rel:
            return None

        # strip the rel prefix from the param name.
        prefix = '%s%s' % (rel or '', LOOKUP_SEP)
        if rel and param.startswith(prefix):
            param = param[len(prefix):]

        # Attempt to match against filters with lookups first. (username__endswith)
        if param in cls.base_filters:
            return param

        # Attempt to match against exclusion filters
        if param[-1] == '!' and param[:-1] in cls.base_filters:
            return param[:-1]

        # Match against relationships. (author__username__endswith).
        # Preference more specific filters. eg, `note__author` over `note`.
        for name in sorted(cls.related_filters, reverse=True):
            # we need to match against '__' to prevent eager matching against
            # like names. eg, note vs note2. Exact matches are handled above.
            if param.startswith("%s%s" % (name, LOOKUP_SEP)):
                return name

    def get_request_filters(self):
        """Build a set of filters based on the request data.

        This currently includes only filter exclusion/negation.

        Returns:
            Mapping of expanded ``{filter names: filter instances}``.
        """
        # build the compiled set of all filters
        requested_filters = OrderedDict()
        for filter_name, f in self.filters.items():
            requested_filters[filter_name] = f

            # exclusion params
            exclude_name = '%s!' % filter_name
            if related(self, exclude_name) in self.data:
                # deepcopy the *base* filter to prevent copying of model & parent
                f_copy = copy.deepcopy(self.base_filters[filter_name])
                f_copy.parent = f.parent
                f_copy.model = f.model
                f_copy.exclude = not f.exclude

                requested_filters[exclude_name] = f_copy

        return requested_filters

    def get_related_filtersets(self):
        """Get the related filterset instances for all related filters.

        Returns:
            Mapping of related ``{filter names: filterset instances}``.
        """
        related_filtersets = OrderedDict()

        for related_name in self.related_filters:
            if related_name not in self.filters:
                continue

            f = self.filters[related_name]
            related_filtersets[related_name] = f.filterset(
                data=self.data,
                queryset=f.get_queryset(self.request),
                relationship=related(self, related_name),
                request=self.request,
                prefix=self.form_prefix,
            )

        return related_filtersets

    def filter_queryset(self, queryset):
        queryset = super(FilterSet, self).filter_queryset(queryset)
        queryset = self.filter_related_filtersets(queryset)
        return queryset

    def filter_related_filtersets(self, queryset):
        """Filter the provided ``queryset`` by the ``related_filtersets``.

        Override this method to change the filtering behavior across relationships.

        Args:
            queryset: The filterset's filtered queryset.

        Returns:
            The ``queryset`` filtered by its related filtersets' querysets.
        """
        for related_name, related_filterset in self.related_filtersets.items():
            # Related filtersets should only be applied if they had data.
            prefix = '%s%s' % (related(self, related_name), LOOKUP_SEP)
            if not any(value.startswith(prefix) for value in self.data):
                continue

            field = self.filters[related_name].field
            to_field_name = getattr(field, 'to_field_name', 'pk') or 'pk'

            field_name = self.filters[related_name].field_name
            lookup_expr = LOOKUP_SEP.join([field_name, 'in'])

            subquery = related_filterset.qs.values(to_field_name)
            queryset = queryset.filter(**{lookup_expr: subquery})

            # handle disinct
            if self.related_filters[related_name].distinct:
                queryset = queryset.distinct()

        return queryset

    def get_form_class(self):
        class Form(super(FilterSet, self).get_form_class()):
            def add_prefix(form, field_name):
                field_name = related(self, field_name)
                return super(Form, form).add_prefix(field_name)

            def clean(form):
                cleaned_data = super(Form, form).clean()

                # when prefixing the errors, use the related filter name,
                # which is relative to the parent filterset, not the root.
                for related_filterset in self.related_filtersets.values():
                    for key, error in related_filterset.form.errors.items():
                        self.form.errors[related(related_filterset, key)] = error

                return cleaned_data
        return Form

    @property
    def form(self):
        from django_filters import compat

        form = super().form
        if compat.is_crispy():
            from crispy_forms.helper import FormHelper

            form.helper = FormHelper(form)
            form.helper.form_tag = False
            form.helper.disable_csrf = True
            form.helper.template_pack = 'bootstrap3'

        return form
